/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Query Interface API
 * 
 * Copyright (C) 2004, Rob Gagnon
 * Rob Gagnon <rgagnon@networkip.net
 *
 * based on code originally written by:
 * Copyright (C) 1999, Mark Spencer
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 * Includes code and algorithms from the Zapata library.
 *
 */

#ifndef _QINT_H
#define _QINT_H

#include <asterisk/channel.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <asterisk/pbx.h>
#include <asterisk/channel.h>
#include <asterisk/chan_sip.h>
#include <asterisk/app_voicemail.h>

// Query Interface Error Types
#define QINT_ERROR	-1
#define QINT_OK		 0

// Query Interface query Types
#define QINT_EXTENSION_LOOKUP	1
#define QINT_SIP_FIND_USER		2
#define QINT_SIP_FIND_PEER		3
#define QINT_SIP_UPDATE_PEER	4
#define QINT_VM_CHANGE_PASSWORD		5
#define QINT_VM_RESET_PASSWORD		6
#define QINT_VM_FIND_USER		7
#define QINT_END_OF_CALL		8

int
load_qint(void);

// returns -1 if something is registered for a given query type
int
qint_isqtype_registered(int qtype);

//! Validate Query Type
/*!
 * \param qtype int to validate as a good Query Interface Type
 * Used to validate a query type
 * Returns -1 if valid, 0 otherwise
 */
int
qint_isvalid_qtype(int qtype);

//! Convert a query type to String
/*!
 * \param qtype int to get the name of
 * Used to get a human-readable version of a query type
 * Returns pointer to character string related to query type
 */
char *
qint_qtype2str(int qtype);

//! Register a Query Handler
/*!
 * \param name name associated with the particular Query Handler
 * \param desc description of the Query Handler
 * \param qtype A query interface query type
 * \param handler function pointer to a Query Handler
 * Used to register a Query handler.
 * Returns QINT_ERROR on error, QINT_OK on success.
 */
extern int
qint_register(char *name, char *desc, int qtype, void *handler);

//! Unregister a Query handler
/*!
 * \param name name of Query handler to unregister
 * \param qtype A query interface query type.  If 0, unregisters all handlers for the handler
 * Unregisters a Query Handler by it's name and type
 */
extern void
qint_unregister(char *name, int qtype);

extern void
qint_unregister_all(char *name);

/**** Definitions for QINT_EXTENSION_LOOKUP
 ******************************************/
typedef int (*qint_exten_lookup_handler)(struct ast_exten *e, struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int action, int *status, char **data);
typedef int (*qint_sip_find_user_handler)(struct sip_user *u, char *user);
typedef int (*qint_sip_find_peer_handler)(struct sip_peer *p, char *peer, struct sockaddr_in *sin);
typedef int (*qint_sip_update_peer_handler)(char *peer, struct sockaddr_in *sin, char *username, int expiry);
typedef int (*qint_vm_change_password_handler)(struct ast_vm_user *vmu, char *password);
typedef int (*qint_vm_reset_password_handler)(char *context, char *mailbox, char *password);
typedef int (*qint_vm_find_user_handler)(struct ast_vm_user *ivm, char *context, char *mailbox);
typedef int (*qint_end_of_call_handler)(struct ast_channel *chan);

//! Send extension information to Extension Lookup Handler for processing
/*!
 * \param chan Channel call is on, or NULL if no live channel
 * Sends information to Extension Lookup Handler for processing
 * Returns a Query Interface Error Type
 */
extern int
qint_exten_lookup(struct ast_exten **e, struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int action, int *status, char **data);

//! Look for a SIP user outside of sip.conf
/*!
 * \param u Address of a struct sip_user pointer. Memory will be allocated and data returned here.
 * \param user String name of SIP user to find
 * Searches for a user and his/her properties using an external source
 * Returns a Query Interface Error Type
 */
extern int
qint_sip_find_user(struct sip_user **u, char *user);

//! Look for a SIP peer outside of sip.conf
/*!
 * \param p Address of a struct sip_peer pointer. Memory will be allocated and data returned here.
 * \param peer String name of SIP peer to find, can be NULL if sin is filled in
 * \param sin Socket information if peer NULL
 * Searches for a peer and his/her properties using an external source
 * Returns a Query Interface Error Type
 */
extern int
qint_sip_find_peer(struct sip_peer **p, char *peer, struct sockaddr_in *sin);

//! Update a peer's locating information in an external source
/*!
 * \param peer String name of SIP peer to update
 * \param sin Socket information (ip and port)
 * \param username
 * \param expiry Timestamp when this registration expires
 * Updates a peer's IP and port number in external source
 * Returns a Query Interface Error Type
 */
extern int
qint_sip_update_peer(char *peer, struct sockaddr_in *sin, char *username, int expiry);

extern int
qint_vm_change_password(struct ast_vm_user *vmu, char *password);

extern int
qint_vm_reset_password(char *context, char *mailbox, char *password);

extern int
qint_vm_find_user(struct ast_vm_user **ivm, char *context, char *mailbox);

extern int
qint_end_of_call(struct ast_channel *chan);

#endif /* _QINT_H */
