/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Query Interface API
 * 
 * Copyright (C) 2004, Rob Gagnon
 * Rob Gagnon <rgagnon@networkip.net>
 *
 * based on code by:
 * Mark Spencer <markster@linux-support.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License.
 *
 * Includes code and algorithms from the Zapata library.
 *
 */

#include <asterisk/lock.h>
#include <asterisk/channel.h>
#include <netinet/in.h>
#include <asterisk/chan_sip.h>
#include <asterisk/app_voicemail.h>
#include <asterisk/qint.h>
#include <asterisk/logger.h>
#include <asterisk/options.h>
#include <asterisk/cli.h>
// #include <asterisk/callerid.h>
// #include <asterisk/causes.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

// Query Interface Handler Item and Mutex Lock
static ast_mutex_t qint_lock = AST_MUTEX_INITIALIZER;
union qint_handler {
	void *fptr;	//generic entry that function pointer is dropped into on register
	qint_exten_lookup_handler exten_lookup;
	qint_sip_find_user_handler sip_find_user;
	qint_sip_find_peer_handler sip_find_peer;
	qint_sip_update_peer_handler sip_update_peer;
	qint_vm_change_password_handler vm_change_password;
	qint_vm_reset_password_handler vm_reset_password;
	qint_vm_find_user_handler vm_find_user;
	qint_end_of_call_handler end_of_call;
};

static struct qint_handleritem {
	char name[20];
	char desc[80];
	int  qtype;
	union qint_handler handler;
	struct qint_handleritem *next;
} *handlers = NULL;

static int qint_cli_show_handlers(int fd, int argc, char *argv[]) {
	struct qint_handleritem *hitem;

	if (handlers) {
		ast_cli(fd,"List of Query Interface Handlers:\n");
	} else {
		ast_cli(fd,"There are no registered Query Interface Handlers.\n\n");
		return RESULT_SUCCESS;
	}
	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		ast_cli(fd, "Query Interface '%s' has handler for %s\n == %s\n",
				hitem->name, qint_qtype2str(hitem->qtype), hitem->desc);
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);
	if (handlers) {
		ast_cli(fd,"=== End of List ===\n\n");
	}
	return RESULT_SUCCESS;
}

static struct ast_cli_entry qint_show_handlers_cli = 
{ {"show", "qint", "handlers", NULL},
	qint_cli_show_handlers, "Shows registered Query Interface Handlers",
	"Shows registered Query Interface Handlers and their descriptions\n"};

int load_qint(void);
int
load_qint(void) {
	ast_cli_register(&qint_show_handlers_cli);
	return 0;
}

int
qint_isqtype_registered(int qtype) {
	struct qint_handleritem *hitem = NULL;

	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == qtype) {
			break;
		}
		hitem = hitem->next;
	}
	if (hitem) {
		ast_verbose(VERBOSE_PREFIX_2 "found registered qtype=%s\n",
			qint_qtype2str(qtype));
		return(-1);
	}
	else {
		return(0);
	}
}

int
qint_isvalid_qtype(int qtype) {
	switch (qtype) {
		// cases that are true / valid.
		// list each QINT_xxxx prior to return -1
		case QINT_EXTENSION_LOOKUP:
		case QINT_SIP_FIND_USER:
		case QINT_SIP_FIND_PEER:
		case QINT_SIP_UPDATE_PEER:
		case QINT_VM_CHANGE_PASSWORD:
		case QINT_VM_RESET_PASSWORD:
		case QINT_VM_FIND_USER:
		case QINT_END_OF_CALL:
			return -1;
			break;
		default:
			return 0;
			break;
	}
}

char *
qint_qtype2str(int qtype) {
	switch (qtype) {
		// Query type translations
		// One case entry for each QINT_xxxx
		case QINT_EXTENSION_LOOKUP:
			return "QINT_EXTENSION_LOOKUP";
			break;
		case QINT_SIP_FIND_USER:
			return "QINT_SIP_FIND_USER";
			break;
		case QINT_SIP_FIND_PEER:
			return "QINT_SIP_FIND_PEER";
			break;
		case QINT_SIP_UPDATE_PEER:
			return "QINT_SIP_UPDATE_PEER";
			break;
		case QINT_VM_CHANGE_PASSWORD:
			return "QINT_VM_CHANGE_PASSWORD";
			break;
		case QINT_VM_RESET_PASSWORD:
			return "QINT_VM_RESET_PASSWORD";
			break;
		case QINT_VM_FIND_USER:
			return "QINT_VM_FIND_USER";
			break;
		case QINT_END_OF_CALL:
			return "QINT_END_OF_CALL";
			break;
		default:
			return "UNDEFINED";
			break;
	}
}

int
qint_register(char *name, char *desc, int qtype, void *handler)
{
	struct qint_handleritem *hitem;

	// Validate parameters
	if (!name)
		return QINT_ERROR;
	if (!qint_isvalid_qtype(qtype)) {
		ast_log(LOG_WARNING, "Query Interface %s, Query Type %d Undefined\n",
				name, qtype);
		return QINT_ERROR;
	}
	if (!handler) {
		ast_log(LOG_WARNING, "Query Interface %s, Query Type %s lacks handler\n",
				name, qint_qtype2str(qtype));
		return QINT_ERROR;
	}

	// Check for duplicate handler
	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == qtype)
			break;
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);
	if (hitem) {
		ast_log(LOG_WARNING, "Already have Query Type %s registered to handler %s.\n", qint_qtype2str(qtype), hitem->name);
		return QINT_ERROR;
	}

	// Allocate memory and setup struct
	hitem = malloc(sizeof(struct qint_handleritem));
	if (!hitem)
		return QINT_ERROR;
	memset(hitem, 0, sizeof(struct qint_handleritem));
	strncpy(hitem->name, name, sizeof(hitem->name) - 1);
	strncpy(hitem->desc, desc, sizeof(hitem->desc) - 1);
	hitem->qtype = qtype;
	hitem->handler = (union qint_handler)handler;
	ast_mutex_lock(&qint_lock);
	hitem->next = handlers;
	handlers = hitem;
	ast_mutex_unlock(&qint_lock);
	ast_verbose(VERBOSE_PREFIX_2 "Query Interface: registered %s for query type %s\n",
				name, qint_qtype2str(qtype));
	return QINT_OK;
}

void
qint_unregister(char *name, int qtype)
{
	struct qint_handleritem *hitem, *prev = NULL;

	if (!name) {
		ast_log(LOG_WARNING, "Missing name attempting to unregister a Query Interface handler.\n");
		return;
	}

	if (!qint_isvalid_qtype(qtype)) {
		ast_log(LOG_WARNING, "Attempt to unregister invalid query type %d from %s Query Interface handler\n",
				qtype, name);
		return;
	}

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (!strcasecmp(name, hitem->name) && (hitem->qtype == qtype)) {
			if (prev)
				prev->next = hitem->next;
			else
				handlers = hitem->next;
			ast_verbose(VERBOSE_PREFIX_2
				"Query Interface: %s unregistered query type %s\n",
				name, qint_qtype2str(hitem->qtype));
			break;
		}
		prev = hitem;
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);
	if (hitem)
		free(hitem);
}

void
qint_unregister_all(char *name) {
		ast_verbose(VERBOSE_PREFIX_2
			"Unregistering all known %s Query Interface handlers\n",
			name);
		qint_unregister(name, QINT_EXTENSION_LOOKUP);
		qint_unregister(name, QINT_SIP_FIND_USER);
		qint_unregister(name, QINT_SIP_FIND_PEER);
		qint_unregister(name, QINT_SIP_UPDATE_PEER);
		qint_unregister(name, QINT_VM_CHANGE_PASSWORD);
		qint_unregister(name, QINT_VM_RESET_PASSWORD);
		qint_unregister(name, QINT_VM_FIND_USER);
		qint_unregister(name, QINT_END_OF_CALL);
}

int
qint_exten_lookup(struct ast_exten **e, struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int action, int *status, char **data)
{
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_EXTENSION_LOOKUP) {
			*e = (struct ast_exten *)malloc(sizeof(struct ast_exten));
			if (!*e) {
				// out of memory
				errno = ENOMEM;
				return QINT_ERROR;
			}
			memset(*e, 0, sizeof(struct ast_exten));
			(*e)->temponly = 1;

			retval = hitem->handler.exten_lookup(*e, chan, context, exten,
							priority, callerid, action, status, data);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_sip_find_user(struct sip_user **u, char *user) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_SIP_FIND_USER) {
			*u = (struct sip_user *)malloc(sizeof(struct sip_user));
			if (!*u) {
				// out of memory
				errno = ENOMEM;
				return QINT_ERROR;
			}
			memset(*u, 0, sizeof(struct sip_user));
			(*u)->temponly = 1;

			retval = hitem->handler.sip_find_user(*u, user);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_sip_find_peer(struct sip_peer **p, char *peer, struct sockaddr_in *sin) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_SIP_FIND_PEER) {
			*p = (struct sip_peer *)malloc(sizeof(struct sip_peer));
			if (!*p) {
				// out of memory
				errno = ENOMEM;
				return QINT_ERROR;
			}
			memset(*p, 0, sizeof(struct sip_peer));
			(*p)->temponly = 1;

			retval = hitem->handler.sip_find_peer(*p, peer, sin);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_sip_update_peer(char *peer, struct sockaddr_in *sin, char *username, int expiry) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_SIP_UPDATE_PEER) {
			retval = hitem->handler.sip_update_peer(peer, sin, username, expiry);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_vm_change_password(struct ast_vm_user *vmu, char *password) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_VM_CHANGE_PASSWORD) {
			retval = hitem->handler.vm_change_password(vmu, password);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_vm_reset_password(char *context, char *mailbox, char *password) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_VM_RESET_PASSWORD) {
			retval = hitem->handler.vm_reset_password(context, mailbox, password);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_vm_find_user(struct ast_vm_user **ivm, char *context, char *mailbox) {
	struct qint_handleritem *hitem;
	struct ast_vm_user *vmu = NULL;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_VM_FIND_USER) {
			// if pointer sent in, use it, instead of allocating one
			if (*ivm) {
				vmu = *ivm;
				memset(vmu, 0, sizeof(struct ast_vm_user));
			}
			else {
				vmu = malloc(sizeof(struct ast_vm_user));
				if (!vmu) {
					// out of memory
					errno = ENOMEM;
					return QINT_ERROR;
				}
				memset(vmu, 0, sizeof(struct ast_vm_user));
				vmu->alloced = 1;
				*ivm = vmu;
			}

			retval = hitem->handler.vm_find_user(vmu, context, mailbox);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}

int
qint_end_of_call(struct ast_channel *chan) {
	struct qint_handleritem *hitem;
	int retval = QINT_OK;

	ast_mutex_lock(&qint_lock);
	hitem = handlers;
	while (hitem) {
		if (hitem->qtype == QINT_END_OF_CALL) {
			retval = hitem->handler.end_of_call(chan);
			break;
		}
		hitem = hitem->next;
	}
	ast_mutex_unlock(&qint_lock);

	return retval;
}
